open Lwt.Infix
open Base

let to_okweb_req conn creq cbody =
  let client_addr = match Conduit_lwt_unix.endp_of_flow conn with
    | `TCP (ipaddr, port) ->
      let inet_addr = match ipaddr with
        | Ipaddr.V4 v4addr -> Ipaddr_unix.V4.to_inet_addr v4addr
        | Ipaddr.V6 v6addr -> Ipaddr_unix.V6.to_inet_addr v6addr
      in
      Some (Unix.ADDR_INET (inet_addr, port))
    | _ -> None
  in
  let http_method = creq |> Cohttp.Request.meth |> Cohttp.Code.string_of_method in
  let uri = creq |> Cohttp.Request.uri in
  let headers = creq |> Cohttp.Request.headers |> Cohttp.Header.to_list |> Headers.of_list in
  (* Last time I checked (2020/03/06) cohttp-lwt-unix ALWAYS used streams
     for request body. So let's skip the checks and also use streams here. *)
  let body = cbody |> Cohttp_lwt.Body.to_stream |> Body.of_stream in
  Request.({
      client_addr = client_addr;
      http_method = http_method;
      uri = uri;
      headers = headers;
      body = body;
    })

let to_cohttp_resp resp =
  let status = resp |> Response.get_status |> Cohttp.Code.status_of_code in
  let headers = resp |> Response.get_headers |> Headers.to_list |> Cohttp.Header.of_list in
  let body = resp |> Response.get_body in
  let (cbody, encoding) = match Body.to_string body with
    | Some str ->
      ( str |> Cohttp_lwt.Body.of_string
      , Cohttp.Transfer.Fixed (String.length str |> Int64.of_int)
      )
    | None ->
      ( body |> Body.to_stream |> Cohttp_lwt.Body.of_stream
      , Cohttp.Transfer.Chunked
      )
  in
  let cresp = Cohttp.Response.make ~status ~headers ~encoding () in
  Lwt.return (cresp, cbody)

let drain_body cbody =
  (* This ia an alternative function for draining the body.

     The original one uses [Lwt_stream.junk_while]. A simple test of
     uploading 2GB file shows that OCaml allocates up to 4GB of memory simply
     to drain the body! I believe it's because [junk_while] reads and
     allocates each piece of a stream because it needs to check the condition.

     This alternative uses [njunk] to unconditionally discard as much of the
     body as possible until the stream is empty. The same test as above now
     shows no excessive allocation of memory. *)
  let stream = Cohttp_lwt.Body.to_stream cbody in
  let rec junk_all () = Lwt_stream.is_empty stream >>= function
    | true -> Lwt.return ()
    | false -> Lwt_stream.njunk (Int.max_value) stream >>= junk_all
  in
  junk_all ()

let default_on_exn ex =
  Caml.Printf.eprintf "Exception occurred: %s\n" (Exn.to_string ex)

let listen ?(on_exn=default_on_exn) ~address ~port handler =
  let callback (conn, _) creq cbody =
    Lwt.finalize
      (fun () -> handler (to_okweb_req conn creq cbody) >>= to_cohttp_resp)
      (fun () -> drain_body cbody)
  in
  let server = Cohttp_lwt_unix.Server.make ~callback () in
  Conduit_lwt_unix.init ~src:address () >>= fun conduit_ctx ->
  let ctx = Cohttp_lwt_unix.Net.init ~ctx:conduit_ctx () in
  let mode = `TCP (`Port port) in
  Cohttp_lwt_unix.Server.create ~on_exn ~ctx ~mode server
