open Alcotest
open Base
open Okweb

let test_of_status_sets_correct_status =
  test_case "of_status sets correct status code" `Quick begin fun () ->
    List.iter [100; 200; 201; 301; 400; 401; 403; 404; 409; 500; ] ~f:begin fun code ->
      check int
        (Printf.sprintf "invalid response status for code %d" code)
        code
        Response.(get_status (of_status code))
    end
  end

let test_with_header_replaces_previous_value =
  test_case "with_header replaces previous value" `Quick begin fun () ->
    let resp1 = Response.(ok |> with_header ~key:"X-Test-1" ~value:"Test Value 1") in
    let resp2 = Response.(resp1 |> with_header ~key:"X-Test-1" ~value:"Test Value 2") in
    let resp3 = Response.(resp1 |> with_header ~key:"X-Test-2" ~value:"Test Value 3") in
    check (option string)
      "invalid value for X-Test-1 in the 1st response"
      (Some "Test Value 1")
      (Headers.find (Response.get_headers resp1) ~key:"X-Test-1");
    check (option string)
      "invalid value for X-Test-1 in the 2nd response"
      (Some "Test Value 2")
      (Headers.find (Response.get_headers resp2) ~key:"X-Test-1");
    check (option string)
      "invalid value for X-Test-2 in the 3rd response"
      (Some "Test Value 3")
      (Headers.find (Response.get_headers resp3) ~key:"X-Test-2");
    check (option string)
      "invalid value for X-Test-1 in the 3rd response"
      (Some "Test Value 1")
      (Headers.find (Response.get_headers resp3) ~key:"X-Test-1");
  end

let test_with_string_body_updates_body_and_sets_content_length =
  test_case "with_string_body updates body and sets Content-Length" `Quick begin fun () ->
    let resp = Response.(ok |> with_string_body "Test") in
    check string
      "invalid body"
      "Test"
      (Response.get_body resp |> Body.to_string |> Option.value ~default:"");
    check (option string)
      "invalid Content-Length"
      (Some "4")
      (Response.get_headers resp |> Headers.find ~key:"Content-Length")
  end

let test_redirect_set_status_headers_and_body =
  test_case "redirect sets correct status, headers and body" `Quick begin fun () ->
    let url = "https://example.com" in
    let expected_body = "<a href=\"https://example.com\">Moved Permanently</a>" in
    let resp = Response.redirect ~url in

    check int
      "invalid status"
      301
      (Response.get_status resp);
    check (option string)
      "invalid Location header"
      (Some url)
      (Response.get_headers resp |> Headers.find ~key:"Location");
    check (option string)
      "invalid Content-Type header"
      (Some "text/html; charset=utf-8")
      (Response.get_headers resp |> Headers.find ~key:"Content-Type");
    check string
      "invalid body"
      expected_body
      (Response.get_body resp |> Body.to_string |> Option.value ~default:"");
  end

let cases = [
  test_of_status_sets_correct_status;
  test_with_header_replaces_previous_value;
  test_with_string_body_updates_body_and_sets_content_length;
  test_redirect_set_status_headers_and_body;
]
